/*
 * voshtomoc.c
 * gaim plugin
 *
 * This a plugin for GAIM to run shell commands through an IM window
 *
 * Copyright (C) 2006 Ioan A. Sucan (isucan@rice.edu)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; This document is under the scope of
 * the version 2 of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define GAIM_PLUGINS

#include <gaim/connection.h>
#include <gaim/conversation.h>
#include <gaim/core.h>
#include <gaim/cmds.h>
#include <gaim/debug.h>
#include <gaim/signals.h>
#include <gaim/util.h>
#include <gaim/notify.h>
#include <gaim/gtkimhtml.h>
#include <gaim/log.h>
#include <gaim/version.h>

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>

/* for strerror_r to be XSI-compliant */
#define _XOPEN_SOURCE 600

/**********************************************************/
/* take differences between gaim 1.5 and 2 into account   */

#if GAIM_MAJOR_VERSION==1 
#define VOSHTOMOC_FOR_OLD_GAIM
#define CONV_TYPE GAIM_CONV_IM
#else
#include <gaim/gtkplugin.h>
#define CONV_TYPE GAIM_CONV_TYPE_IM
#endif

/**********************************************************/

#define PNAME "voshtomoc"
#define PVERSION "0.2"
static const int MAX_MSG_SIZE = 448;      /* data we send */

/* data for programs we may run */

typedef struct {
    char  *name;          /* the name of the program to run    */
    int    n_allowed;     /* number of allowed users           */
    char **allowed;       /* the allowed users                 */
    int    is_cmd;        /* parse as command (!CMD[...]) if 1 */
    int    send_name;     /* send the user name as well if 1   */
    int    when_away;     /* active only when away if 1        */
} program_t;

static int        n_progs = 0;         /* number of available programs */
static program_t *progs   = NULL;      /* list of available programs   */

/**********************************************************/

/* data for active users */

typedef struct {
    char *name;      /* the alias for the user in question (what is displayed by gaim) */
    int rwpipe[2];   /* the pipes used for communicating with child process; write to 1, read from 0 */
    pid_t pid;       /* the pid of the child process */
    int prog;        /* the number of the program_t structure used to run the child process */
} user_t;

static user_t *users   = NULL;    /* the list of users */
static int     n_users = 0;       /* the number of users */

/**********************************************************/

/* constants for command processing */
static const char *CMD       = "!CMD";     /* command marker */

static const int   SEP_COUNT = 4;          /* number of separator symbols */
static const char *BEG       = "([<{";     /* starting separator */
static const char *END       = ")]>}";     /* terminating separator */


/*********************************************************/

/* registered commands */

GaimCmdId reload_cmd;

/*********************************************************/

/* for every user, we keep an internal ID, to identify the program
 * this user is connected to */
static int get_user_id(char *user, int *add_if_not_found) {
    
    int i;
    gaim_debug(GAIM_DEBUG_INFO, PNAME, "%s for %s\n", __func__, user);
    
    if (add_if_not_found != NULL) *add_if_not_found = 0;
    for (i = 0 ; i < n_users ; i++)
	if (strcmp(user, users[i].name) == 0) return i;

    /* add user if not found */
    if (add_if_not_found != NULL) {
	*add_if_not_found = 1;
	int old_n_users = n_users;
	n_users++;
	users = (user_t*)realloc(users, n_users*sizeof(user_t));
	users[old_n_users].name = strdup(user);
	users[old_n_users].rwpipe[0] = users[old_n_users].rwpipe[1] = -1;
	users[old_n_users].pid = -1;
	users[old_n_users].prog = -1;
	return old_n_users;
    }
    
    return -1;
}

/* remove the data for a specific user */
static void remove_user_id(int id) {

    gaim_debug(GAIM_DEBUG_INFO, PNAME, "%s\n", __func__);
    
    /* optimistic, assume correct id and a succesful removal */
    n_users--;
    if (id < n_users && id >= 0) {
	
	free(users[id].name);
	users[id].name = users[n_users].name;    
	
	kill(users[id].pid, SIGTERM);
	users[id].pid = users[n_users].pid;
	
	close(users[id].rwpipe[0]); close(users[id].rwpipe[1]);
	users[id].rwpipe[0] = users[n_users].rwpipe[0];
	users[id].rwpipe[1] = users[n_users].rwpipe[1];
	
    } else
	if (id == n_users) {
	    free(users[n_users].name);
	    kill(users[n_users].pid, SIGTERM);
	    close(users[n_users].rwpipe[0]); 
	    close(users[n_users].rwpipe[1]);
	}
	else {
	    /* undo the change in number of user if removal fails */
	    n_users++;
	    return;
	}
    users = (user_t*)realloc(users, n_users*sizeof(user_t));
}

/* check whether a specific buddy is allowed to execute some program with command syntax */
static int buddy_allow_cmd(char* user, int is_idle) {
    int i, j;
    for (i = 0 ; i < n_progs ; i++)
	if (progs[i].is_cmd && (!progs[i].when_away || is_idle != 0) && progs[i].name[0] != '#')
	    for (j = 0; j < progs[i].n_allowed; j++)
		if (strcmp(progs[i].allowed[j], "*") == 0 ||
		    strcmp(user, progs[i].allowed[j]) == 0) return i;
    return -1;
}

/* check whether a specific buddy is allowed to execute some program without command syntax */
static int buddy_allow_non_cmd(char* user, int is_idle) {
    int i, j;
    for (i = 0 ; i < n_progs ; i++)    
	if (!progs[i].is_cmd && (!progs[i].when_away || is_idle != 0) && progs[i].name[0] != '#')
	    for (j = 0; j < progs[i].n_allowed; j++)
		if (strcmp(progs[i].allowed[j], "*") == 0 || 
		    strcmp(user, progs[i].allowed[j]) == 0) return i;
    return -1;
}

/*******************************************************/

/* retrieve the command from !CMD(<command>) 
 * caller must free the allocated memory */

static char* get_sent_command(char **buffer) {
    char* cmd = NULL;
    gaim_debug(GAIM_DEBUG_INFO, PNAME, "%s\n", __func__);
    
    if (buffer != NULL && *buffer != NULL) {
	int n = strlen(*buffer);
	int in_parse = 0;
	int in_retr = 0;
	int parse_pos = -4;
	char wait_for = 0;
	int cmd_start = 0;
	int cmd_len = 0;
	int i;
	for (i = 0 ; i < n ; i++) {
	    if ((*buffer)[i] == '!' && !in_parse && !in_retr) {
		in_parse = 1;
		parse_pos = i;
		continue;
	    }
	    if (in_parse)
		if (i-parse_pos > 3) {
		    int j;
		    wait_for = (*buffer)[i];
		    for (j = 0; j < SEP_COUNT; j++) 
			if ((*buffer)[i] == BEG[j]) {
			    wait_for = END[j];
			    break;
			}
		    cmd_start = i + 1;
		    in_parse = 0;
		    in_retr = 1;
		    continue;
		} else
		    if ((*buffer)[i] != CMD[i-parse_pos]) in_parse = 0;
	    if (in_retr)
		if ((*buffer)[i] == wait_for) {
		    cmd_len = i - cmd_start;
		    break;
		}      
	}
	if (cmd_len > 0) {
	    cmd = (char*)malloc((cmd_len+1)*sizeof(char));
	    strncpy(cmd, &((*buffer)[cmd_start]), cmd_len);
	    cmd[cmd_len] = 0;
	}
    }
    return cmd;
}


/* Remove gaim markers */

static char* strip_gaim_tags(char* buffer) {
    int i;
    int j = 0;
    int ignore = 0;
    int n = strlen(buffer);
    char *result = (char*)malloc(sizeof(char)*(n + 1));
    for (i = 0 ; i < n ; i++) {
	if (buffer[i] == '<') ignore = 1; else
	    if (buffer[i] == '>') ignore = 0; else
		if (!ignore) result[j++] = buffer[i];
    }
    result[j] = 0;
    result = (char*)realloc(result, sizeof(char)*j);
    return result;
}


/* Find the process id for a user. If user was not seen before,
 * create the ID, run the application and set up the pipes */

static int get_id_and_setup_child_process(char* sender, int prg_id) {
    int add = 0;
    int id = get_user_id(sender, &add);
  
  /* fork, exec & setup for the pipes */
    if (add) {
	pid_t pid;
	int extra_pipe[2];
	gaim_debug(GAIM_DEBUG_INFO, PNAME, "Running program '%s'\n", progs[prg_id].name);
	pipe(users[id].rwpipe);
	pipe(extra_pipe);	  
	pid = fork();
	if (pid == -1) {
	    close(users[id].rwpipe[0]); close(users[id].rwpipe[1]);
	    close(extra_pipe[0]); close(extra_pipe[1]);
	    users[id].rwpipe[0] = users[id].rwpipe[1] = -1;
	} else 
	    if (pid == 0) { // child
		dup2(users[id].rwpipe[1], 1);
		dup2(users[id].rwpipe[1], 2);
		close(users[id].rwpipe[1]);
		close(users[id].rwpipe[0]);
		
		dup2(extra_pipe[0], 0);
		close(extra_pipe[0]);
		close(extra_pipe[1]);
		
		execlp(progs[prg_id].name, progs[prg_id].name, NULL);
		
		{	  
		    char err_str[512] = {0};
		    strerror_r(errno, err_str, sizeof(err_str));
		    fprintf(stderr, "execlp(%s) failed: %s\nPATH='%s'", 
			    progs[prg_id].name, err_str, getenv("PATH"));
		}
		
		_exit(errno);
		
	    } else {  // parent
		close(users[id].rwpipe[1]);
		close(extra_pipe[0]);
		users[id].rwpipe[1] = extra_pipe[1];
		users[id].pid = pid;
		users[id].prog = prg_id;
		// we now write to pipes[id][1]  which is stdin to child
		// and read from pipes[id][0] which is stdout & stderr to child
	    }
    }
    
    return id;
}

/* send a command to the pipe and get the reply */
static char* get_reply(int id, char* command, char* sender) {
    char temp_buf[16384];
    if (command != NULL) {
	int err;
	memset(temp_buf, 0, sizeof(temp_buf));
	if (progs[users[id].prog].send_name)
	    snprintf(temp_buf, sizeof(temp_buf), "%s\n%s\n", sender, command);
	else
	    snprintf(temp_buf, sizeof(temp_buf), "%s\n", command);
	err = write(users[id].rwpipe[1], temp_buf, strlen(temp_buf));
	memset(temp_buf, 0, sizeof(temp_buf));
	err = read(users[id].rwpipe[0], temp_buf, sizeof(temp_buf));
	return (strcmp(temp_buf, "<empty>\n") == 0) ? NULL : strdup(temp_buf);
    }
    return NULL;
}

/* handler for gaim received messages */
static gboolean message_recv(GaimAccount *account, char **sender, char **buffer, int *flags)
{
    char *reply = NULL;
    
    /* get the buddy alias first */
    GaimBuddy *buddy = (sender != NULL && *sender != NULL) ? gaim_find_buddy(account, *sender) : NULL;
    char* sender_alias = (buddy != NULL) ? (char*)gaim_buddy_get_alias(buddy) : NULL;
    
#ifdef VOSHTOMOC_FOR_OLD_GAIM
    int is_idle = gaim_account_get_connection(account)->is_idle;
#else
    int is_idle = gaim_presence_is_idle(gaim_account_get_presence(account));
#endif

    if (sender_alias != NULL) {
	int skip_rest = 0;
	int the_prg;
	
	/* process commands */
	the_prg = buddy_allow_cmd(*sender, is_idle);
	if (the_prg >= 0) {
	    char *cmd = (char*)get_sent_command(buffer);
	    if (cmd != NULL) {
		int id = get_id_and_setup_child_process(sender_alias, the_prg);
		reply = get_reply(id, cmd, *sender);
		free(cmd);
		skip_rest = 1;
	    }
	}
	
	if (!skip_rest) {
	    the_prg = buddy_allow_non_cmd(*sender, is_idle);
	    if (the_prg >= 0) {
		char *text = (char*)strip_gaim_tags(*buffer);
		if (text != NULL) {
		    int id = get_id_and_setup_child_process(sender_alias, the_prg);
		    reply = get_reply(id, text, *sender);
		    free(text);
		}
	    }
	}
    }
    
    if (reply != NULL) {
	
	GaimConversation *conv = 
	    gaim_find_conversation_with_account(
#ifndef VOSHTOMOC_FOR_OLD_GAIM
						CONV_TYPE,
#endif
						*sender, account);
	
	if (conv == NULL)
	    conv = gaim_conversation_new(CONV_TYPE, account, *sender);
	GaimConvIm *im = (conv != NULL) ? gaim_conversation_get_im_data(conv) : NULL;
	if (im != NULL) { 
	    int l = strlen(reply);
	    int s = 0;
	    while (l - s > MAX_MSG_SIZE) {
		char backup = reply[s + MAX_MSG_SIZE];
		reply[s + MAX_MSG_SIZE] = 0;
		gaim_conv_im_send(im, &(reply[s]));
		s = s + MAX_MSG_SIZE;
		reply[s] = backup;
	    }
	    if (l > s)
		gaim_conv_im_send(im, &(reply[s]));
	}
	free(reply);
    }
    
    return FALSE;
}

/* handler for gaim sign off. We close potentially started programs */
static void buddy_sign_off(GaimBuddy *buddy) {
    int buddy_id = get_user_id((char*)gaim_buddy_get_alias(buddy), NULL);
    if (buddy_id >= 0) remove_user_id(buddy_id);
}


/* load the plugin configuration file */
static void load_plugin_config() {
    FILE *file = fopen("/etc/voshtomoc.conf", "r");
    
    if (file == NULL) {
	char home[512];
	snprintf(home, sizeof(home), "%s/.gaim/voshtomoc.conf", getenv("HOME"));
	file = fopen(home, "r");
    }
    
    if (file != NULL) {
	char buffer[4096];
	int i, j;
	
	fscanf(file, "%d", &n_progs);
	progs = (program_t*)malloc(n_progs*sizeof(program_t));
	for (i = 0; i < n_progs; i++) {
	    fscanf(file, "%s", buffer);
	    progs[i].name = strdup(buffer);
	    fscanf(file, "%d", &(progs[i].is_cmd));
	    fscanf(file, "%d", &(progs[i].send_name));
	    fscanf(file, "%d", &(progs[i].when_away));
	    fscanf(file, "%d", &(progs[i].n_allowed));
	    progs[i].allowed = (char**)malloc(progs[i].n_allowed*sizeof(char*));
	    for (j = 0 ; j < progs[i].n_allowed; j++) {
		fscanf(file, "%s", buffer);
		progs[i].allowed[j] = strdup(buffer);
	    }
	}
	
	fclose(file);
	
    } else {
	fprintf(stderr, "%s: No configuration file found.\n", PNAME);
	gaim_debug(GAIM_DEBUG_INFO, PNAME, "No configuration file found\n");
    }
}

/* free memory allocated for configuration */
static void free_plugin_config() {
    int i, j;
    for (i = 0 ; i < n_progs ; i++) {
	free(progs[i].name);
	for (j = 0 ; j < progs[i].n_allowed ; j++)
	    free(progs[i].allowed[j]);
	free(progs[i].allowed);
    }
    free(progs);
    n_progs = 0; progs = NULL;
}

/* close all running programs (children) */
static void close_plugin_pipes() {
    int i;
    for (i = 0; i < n_users; i++) { 
	free(users[i].name);
	kill(users[i].pid, SIGTERM);
	close(users[i].rwpipe[0]);
	close(users[i].rwpipe[1]);
    }
    free(users);
    n_users = 0; users = NULL;
}

/* handler for the \vreload command. We reload the plugin configuration */
static GaimCmdRet reload_plugin_config(GaimConversation* conv, const gchar* cmd,
				       gchar** args, gchar** err, void* data) {
    free_plugin_config();
    close_plugin_pipes();
    load_plugin_config();
    gaim_debug(GAIM_DEBUG_INFO, PNAME, "reloaded configuration\n");
    return GAIM_CMD_RET_OK;
}

/* to be run when loading the plugin */
static gboolean plugin_load(GaimPlugin *plugin)
{
    void *conv_handle = gaim_conversations_get_handle();
    
    gaim_signal_connect(conv_handle, "receiving-im-msg",
			plugin, GAIM_CALLBACK(message_recv), NULL);
    
    void *blist_handle = gaim_blist_get_handle();
    
    gaim_signal_connect(blist_handle, "buddy-signed-off",
			plugin, GAIM_CALLBACK(buddy_sign_off), NULL);
    
    reload_cmd = gaim_cmd_register("vreload", "", GAIM_CMD_P_DEFAULT, GAIM_CMD_FLAG_IM,
				   NULL, reload_plugin_config, "vreload:  Reload the configuration file for voshtomoc.", NULL);
    
    load_plugin_config();
    
    gaim_debug(GAIM_DEBUG_INFO, PNAME, "%s loaded\n", PNAME);
    
    return TRUE;
}

/* free all allocated memory */
static void free_plugin_memory() {
    free_plugin_config();
    close_plugin_pipes();
}


/* to be run when unloading the plugin */
static gboolean plugin_unload(GaimPlugin * plugin) {
    void *conv_handle = gaim_conversations_get_handle();
    gaim_signal_disconnect(conv_handle, "receiving-im-msg", plugin, GAIM_CALLBACK(message_recv));
    
    void *blist_handle = gaim_blist_get_handle();
    gaim_signal_disconnect(blist_handle, "buddy-signed-off", plugin, GAIM_CALLBACK(buddy_sign_off));
    
    gaim_cmd_unregister(reload_cmd);
    
    free_plugin_memory();
    
    gaim_debug(GAIM_DEBUG_INFO, PNAME, "%s unloaded\n", PNAME);
    
    return TRUE;
}

#ifndef VOSHTOMOC_FOR_OLD_GAIM

static GaimPluginPrefFrame* get_plugin_pref_frame(GaimPlugin *plugin) {
    GaimPluginPrefFrame *frame;
    frame = gaim_plugin_pref_frame_new();
    
    
    
    return frame;
}

/* preferences for voshtomoc */
static GaimPluginUiInfo prefs_info = {
    get_plugin_pref_frame
};

#endif

/* plugin information */
static GaimPluginInfo info =
  {
      GAIM_PLUGIN_MAGIC,
      GAIM_MAJOR_VERSION,
      GAIM_MINOR_VERSION,
      GAIM_PLUGIN_STANDARD,                             /**< type           */
      NULL,                                             /**< ui_requirement */
      0,                                                /**< flags          */
      NULL,                                             /**< dependencies   */
      GAIM_PRIORITY_DEFAULT,                            /**< priority       */
      
      "_voshtomoc",                                     /**< id             */
      PNAME,                                            /**< name           */
      PVERSION,                                         /**< version        */
      /**  summary        */
      "Run external applications and use their STDIN/STDOUT in IMs.",
      /**  description    */
      "If you have a program that you interact with using STDIN/STDOUT, you can run it with this plugin and have a buddy interact with your program using am IM window. You can specify the interaction to occur with only specific buddies, for all messages or for commands only (specified with a message of the form !CMD[command]), or restrict the interaction to occur only when you are marked as away.",
      "Ioan A. Sucan <isucan@rice.edu>",                /**< author         */
      "http://www.robodesign.ro/isucan/other",          /**< homepage       */
      plugin_load,                                      /**< load           */
      plugin_unload,                                    /**< unload         */
      NULL,                                             /**< destroy        */
      NULL,                                             /**< ui_info        */
      NULL,                                             /**< extra_info     */
      
#ifdef VOSHTOMOC_FOR_OLD_GAIM
      NULL,
#else
      &prefs_info,
#endif
      
      NULL
  };


/* plugin initialization */
static void
init_plugin(GaimPlugin *plugin)
{
}

GAIM_INIT_PLUGIN(voshtomoc, init_plugin, info)
